/*----------------------------------------------------------------------+
|									|
|   Copyright (C) 1991   Bill Steinbock, All rights reserved.           |
|									|
|   datstmp.mc -  MDL program to place date/file stamp.  This program   |
|                 places text showing the design file name and date.    |
|                 This program support the following commands:          |
|                                                                       |
|                 PLACE STAMP  - places text                            |
|                 UPDATE STAMP - updates text previously placed         |
|                                                                       |
|                                                                       |
|  -------------------------------------------------------------------  |
|                                                                       |
|   cadgurus.com has made the following changes to datstmp.mc			|
|                                                                       |
|   added placement dialog to override attributes: color, weight, level |
|   line style, text height, text width, text font, and placement angle |
|                                                                       |
|                                                                       |
|																		|
+----------------------------------------------------------------------*/
/*----------------------------------------------------------------------+
|									|
|   Include Files   							|
|									|
+----------------------------------------------------------------------*/
#include    <mdl.h>
#include    <global.h>
#include    <mselems.h>
#include    <userfnc.h>
#include    <rscdefs.h>
#include    <scanner.h>
#include    <tcb.h>
#include    <time.h>
#include    <string.h>


#include    "datstmp.h"
#include    "datstcmd.h"
#include    "fdf.fdf"


/*----------------------------------------------------------------------+
|									|
|    Local defines                                                      |
|									|
+----------------------------------------------------------------------*/
#define     GRAFIC          1160
#define     ACTIVE_FILE     0

Public DatstmpGlobals datstmpGlobals;

Private DialogBox   *dsprefs;

/*----------------------------------------------------------------------+
|									|
| local function declaration                                            |
|									|
+----------------------------------------------------------------------*/
/*----------------------------------------------------------------------+
|																		|
|   Hook Function Declarations											|
|																		|
+----------------------------------------------------------------------*/
Private void textAngle_textItemHook ( DialogItemMessage   *dimP );

Private DialogHookInfo uHooks[] =
    {
    {HOOKITEMID_textAngle,			textAngle_textItemHook},
    };


/*----------------------------------------------------------------------+
|																		|
| name		textAngle_textItemHook 										|
|						  												|
| author	Mark Stefanchuk									1/98		|
|																		|
+----------------------------------------------------------------------*/
Private void textAngle_textItemHook
(
DialogItemMessage   *dimP
)
    {
    dimP->msgUnderstood = TRUE;
    switch (dimP->messageType)
        {
        case DITEM_MESSAGE_CREATE:
            {  
            
            break;
            }

        case DITEM_MESSAGE_SETSTATE:
            {  
            mdlString_toAngle (&datstmpGlobals.angleVal, datstmpGlobals.textAngle);

            break;
            }

        case DITEM_MESSAGE_GETSTATE:
            {                
             
            mdlString_toAngle (&datstmpGlobals.angleVal, datstmpGlobals.textAngle);

            mdlString_fromAngle (datstmpGlobals.textAngle, datstmpGlobals.angleVal, 
                                 tcb->anglefmt, FALSE, -2, TRUE);
             
            break;
            }

        default:
            {
            dimP->msgUnderstood = FALSE;
            break;
            }
        }
    }


/*----------------------------------------------------------------------+
|									|
| name          getDateTime                                             |
|									|
+----------------------------------------------------------------------*/
Private void    getDateTime
(
char    *dateString
)
    {
    char            timeString[6], day[3], month[4], year[3];
    char            dateTime[26];
    time_t          dayTime;

    time (&dayTime);
    strcpy (dateTime, ctime(&dayTime));

    strncpy (timeString,  dateTime+11, 5);
    strncpy (day,         dateTime+ 8, 2);
    strncpy (month,       dateTime+ 4, 3);
    strncpy (year,        dateTime+22, 2);

    timeString[5] = 0;
    day[2]        = 0;
    month[3]      = 0;
    year[2]       = 0;

    sprintf (dateString, "DATE: %s-%s-%s  %s", day, month, year, timeString);
    }




Private int getPoint
(
Dpoint3d		*point,
Dpoint3d		*origin,
double			lineSpacing
)
{
	Dpoint3d	xVec, tmpPoint;
	Transform	tMatrix;

	tmpPoint	= *origin;
	tmpPoint.x 	= origin->x + 1.0;

	mdlTMatrix_getIdentity ( &tMatrix );
	mdlTMatrix_rotateByAngles ( &tMatrix, &tMatrix, 0.0, 0.0, 
						((datstmpGlobals.angleVal  + 90) * fc_piover180) );
	mdlTMatrix_setOrigin ( &tMatrix, origin );
	
	mdlTMatrix_transformPoint ( &tmpPoint, &tMatrix );
	mdlVec_computeNormal ( &xVec, &tmpPoint, origin);
																	
	mdlVec_projectPoint ( point, origin, &xVec, lineSpacing );

	return SUCCESS;
}

/*----------------------------------------------------------------------+
|									|
| name          generateTextElems - function to create text.            |
|									|
+----------------------------------------------------------------------*/
Private int     generateTextElems
(
Dpoint3d    *ptP,
int	    view,
int         drawMode,
int         writeMode
)
    {
	RotMatrix       rMatrix;
    MSElementUnion  el;
    MSElementDescr  *textElmDP;
    Dpoint3d        origin, point;
    char            textString[255];
    TextParam       txtParam;
    double          lineSpacing;
	TextSizeParam   txtSize;
	int				color, weight, style, level;


    txtParam.font            = tcb->actfont;
    txtParam.just            = TXTJUST_LT;
    txtParam.style           = -1;
    txtParam.viewIndependent = FALSE;

	txtSize.size.height =  tcb->chheight;
	txtSize.size.width =   tcb->chwidth;

    mdlCurrTrans_begin();
    mdlCurrTrans_identity();
    mdlCurrTrans_rotateByView (view);
    mdlCurrTrans_invtransPointArray (&origin, ptP, 1);

    /* --- create design file name --- */
    sprintf (textString, "FILE: %s",tcb->dgnfilenm);

	if (datstmpGlobals.usePrefs) 
	{
		txtSize.size.height =  datstmpGlobals.maHeight;
		txtSize.size.width =   datstmpGlobals.maWidth;

		txtParam.font	= datstmpGlobals.maFont;

		// set text justification - single line text only
		switch (datstmpGlobals.maTxtJust)
		{
	   	    case 0:
	            txtParam.just   =  TXTJUST_LT;
				break;
			case 1:
	            txtParam.just   =  TXTJUST_LC;
				break;
			case 2:
	            txtParam.just   =  TXTJUST_LB;
				break;
			case 3:
	            txtParam.just   =  TXTJUST_CT;
				break;
			case 4:
	            txtParam.just   =  TXTJUST_CC;
				break;
			case 5:
	            txtParam.just   =  TXTJUST_CB;
				break;
			case 6:
	            txtParam.just   =  TXTJUST_RT;
				break;
			case 7:
	            txtParam.just   =  TXTJUST_RC;
				break;
			case 8:
	            txtParam.just   =  TXTJUST_RB;
				break;
	        default:
	            txtParam.just   =  TXTJUST_LT;
				break;
	     }   				
	} 

	if (datstmpGlobals.usePrefs)
	{
	    mdlRMatrix_fromAngle (&rMatrix, datstmpGlobals.angleVal * fc_piover180 );
        mdlText_create (&el, NULL, textString, &origin, &txtSize, &rMatrix,
                    &txtParam, NULL);
	}
	else
	{
	     mdlText_create (&el, NULL, textString, &origin, &txtSize, NULL,
                    &txtParam, NULL);
	}

    if (datstmpGlobals.usePrefs)
	{
		mdlElement_setSymbology ( &el, &datstmpGlobals.maColor, 
							&datstmpGlobals.maWeight, &datstmpGlobals.maStyle);
		mdlElement_setProperties ( &el, &datstmpGlobals.maLevel, 
		 					NULL, NULL, NULL, NULL, NULL, NULL, NULL );
    }

    mdlElmdscr_new (&textElmDP, NULL, &el);

    /* --- move down --- */
    point = origin;
    lineSpacing = tcb->chheight * 1.5;
	//lineSpacing = tcb->nodespace;

	if (datstmpGlobals.usePrefs)
	{
	    lineSpacing = datstmpGlobals.lineSpace;
//>>> define point as a normal displacement from origin
		getPoint (&point, &origin, lineSpacing);

	}
	else
    point.y = point.y - lineSpacing;

    /* --- display date and time text --- */
    getDateTime (textString);

	if (datstmpGlobals.usePrefs)
	{
	    mdlRMatrix_fromAngle (&rMatrix, datstmpGlobals.angleVal * fc_piover180 );

        mdlText_create (&el, NULL, textString, &point, &txtSize, &rMatrix,
                    &txtParam, NULL);
	}
	else
	{
	     mdlText_create (&el, NULL, textString, &point, &txtSize, NULL,
                    &txtParam, NULL);
	}

    if (datstmpGlobals.usePrefs)
	{
		mdlElement_setSymbology ( &el, &datstmpGlobals.maColor, 
							&datstmpGlobals.maWeight, &datstmpGlobals.maStyle);
		mdlElement_setProperties ( &el, &datstmpGlobals.maLevel, 
		 					NULL, NULL, NULL, NULL, NULL, NULL, NULL );
    }


    mdlElmdscr_appendElement (textElmDP, &el);

    mdlElmdscr_display (textElmDP, 0, drawMode);

    /* --- actually add text as text node to file if required --- */
    if (writeMode)
        mdlElmdscr_add (textElmDP);

    /* --- free the element descriptor --- */
    mdlElmdscr_freeAll (&textElmDP);

    mdlCurrTrans_end();
    return  SUCCESS;
    }

/*----------------------------------------------------------------------+
|									|
| name          cleanUp                                                 |
|									|
+----------------------------------------------------------------------*/
Private void    cleanUp()
    {
    tcb->cugraf=0;
    mdlCurrTrans_clear();
    mdlState_startDefaultCommand ();
    }

/*----------------------------------------------------------------------+
|									|
| name          updateText                                              |
|									|
+----------------------------------------------------------------------*/
Private int     updateText
(
MSElementUnion  *elP
)		
    {
    char   existingText[255], matchText[8], newText[255];

    if (mdlElement_getType (elP) != TEXT_ELM ||
        mdlText_extractString (existingText, elP))
        return  MODIFY_STATUS_NOCHANGE;
    
    strcpy (matchText, "FILE: ");
    if (!strncmp (matchText, existingText, 6))
        {
        /* --- update the filename --- */
        sprintf (newText, "FILE: %s",tcb->dgnfilenm);
        mdlText_create (elP, elP, newText, NULL, NULL, NULL, NULL, NULL);
        return MODIFY_STATUS_REPLACE;
        }

    strcpy (matchText, "DATE: ");
    if (!strncmp (matchText, existingText, 6))
        {
        /* --- update the date/time --- */
        getDateTime (newText);
        mdlText_create (elP, elP, newText, NULL, NULL, NULL, NULL, NULL);
        return MODIFY_STATUS_REPLACE;
        }

    return MODIFY_STATUS_NOCHANGE;
    }

/*----------------------------------------------------------------------+
|									|
| name          updateStamp                                             |
|									|
+----------------------------------------------------------------------*/
Private void    updateStamp ()
    {
    ULong	    elemAddr[50], eofPos, filePos;
    int             scanWords, status, i, numAddr;
    ExtScanlist	    scanList;

    mdlScan_initScanlist (&scanList);
    mdlScan_noRangeCheck (&scanList);

    scanList.scantype       = ELEMTYPE;
    scanList.extendedType   = FILEPOS;
    scanList.typmask[0]	    = TMSK0_TEXT_NODE;
    scanList.typmask[1]	    = TMSK1_TEXT;
    
    eofPos  = mdlElement_getFilePos (FILEPOS_EOF, NULL);
    filePos = 0L;

    mdlScan_initialize (0, &scanList);

    /* --- loop through all text elements to find FILE: and DATE: --- */
    do
	{
	scanWords = sizeof(elemAddr)/sizeof(short);
	status	  = mdlScan_file (elemAddr, &scanWords, sizeof(elemAddr),
			&filePos);
	numAddr   = scanWords / sizeof(short);

	for (i=0; i<numAddr; i++)
	    {
	    if (elemAddr[i] >= eofPos)
		break;

	    mdlModify_elementSingle (0, elemAddr[i], MODIFY_REQUEST_NOHEADERS,
                        MODIFY_ORIG, updateText, NULL, 0L);
	    }
	} while (status == BUFF_FULL);

    return;
    }


/*----------------------------------------------------------------------+
|									|
| name          generateText - function to show text dynamically        |
|									|
+----------------------------------------------------------------------*/
Private void    generateText
(
Dpoint3d    *pt,
int	    view,
int         drawMode
)
    {
    generateTextElems (pt, view, drawMode, 0);
    }

/*----------------------------------------------------------------------+
|									|
| name          placeText                                               |
|									|
+----------------------------------------------------------------------*/
Private void    placeText
(
Dpoint3d    *pt,
int         view
)
    {
    MSElementUnion  el;
    int             size=1, offset=GRAFIC;

    /* --- set graphic group number --- */
    tcb->cugraf=tcb->graphic;
    tcb->graphic++;

    generateTextElems (pt, view, NORMALDRAW, 1);

    /* --- update type 9 with new GG number --- */
    mdlParams_storeType9Variable (&tcb->graphic, size, offset);

    cleanUp ();
    }

/*----------------------------------------------------------------------+
|									|
| name          placeStampCmd                                           |
|									|
+----------------------------------------------------------------------*/
void            placeStampCmd ()
cmdNumber       CMD_PLACE_STAMP
    {
	// cadgurus.com - make sure dialog is open
	dsprefs = mdlDialog_open (NULL, DIALOGID_prefs);
    /* --- Set datapoint, reset, and dynamics functions --- */
    mdlCurrTrans_clear();
    mdlState_startPrimitive (NULL, cleanUp, 0, 0);
    mdlState_setFunction (STATE_DATAPOINT,        placeText);
    mdlState_setFunction (STATE_COMPLEX_DYNAMICS, generateText);
    mdlOutput_rscPrintf (MSG_PROMPT, NULL, 0, 2);
    }

/*----------------------------------------------------------------------+
|									|
| name          updateStampCmd                                          |
|									|
+----------------------------------------------------------------------*/
void            updateStampCmd ()
cmdNumber       CMD_UPDATE_STAMP
    {
	// cadgurus.com - make sure dialog is open
	dsprefs = mdlDialog_open (NULL, DIALOGID_prefs);
    mdlOutput_rscPrintf (MSG_PROMPT, NULL, 0, 3);
    updateStamp ();
    mdlOutput_prompt (" ");
    mdlState_startDefaultCommand ();
    }

/*----------------------------------------------------------------------+
|																		|
| name		bye function saves application preference data				|
|																		|																		|
+----------------------------------------------------------------------*/
Private int bye ()
   {
   RscFileHandle dataUserPrefs;
   DatstmpGlobals *globalP;
   char cfgVal[128];
   char tmpStr [128];
   char *cfgVarExpansion = NULL;
   int status, i;

   // figure out where MicroStation is saving resource files
   mdlSystem_getCfgVar (cfgVal, "MS_RSRCPATH", sizeof(cfgVal));

   //  Expands the configuration variable
   cfgVarExpansion = mdlSystem_expandCfgVar (cfgVal);

   if (cfgVarExpansion != NULL)
       {
       sprintf (tmpStr, "%s", cfgVarExpansion);
       free (cfgVarExpansion);
       }

   if (tmpStr)
   {
       strcat (tmpStr, "datstmp.rsc");

	   status = mdlResource_openFile (&dataUserPrefs, tmpStr, RSC_READWRITE);

	   if (status != SUCCESS)
	      {
	      mdlResource_createFile (tmpStr, NULL, MAVERSION);
	      mdlResource_openFile (&dataUserPrefs, tmpStr, RSC_READWRITE);
	      }

     	globalP = (DatstmpGlobals *)mdlResource_load (dataUserPrefs,
                              RTYPE_globalPrefs, RSCID_globalPrefs);
     	if (!globalP)
             {
             mdlResource_add (dataUserPrefs, 
                             RTYPE_globalPrefs, RSCID_globalPrefs,
                             &datstmpGlobals, sizeof(DatstmpGlobals), NULL);
             }
    	 else
             {
             globalP->maLevel = datstmpGlobals.maLevel;
             globalP->maWeight = datstmpGlobals.maWeight;
             globalP->maColor = datstmpGlobals.maColor;
             globalP->maStyle = datstmpGlobals.maStyle;
			 globalP->maHeight = datstmpGlobals.maHeight;
			 globalP->maWidth = datstmpGlobals.maWidth;
			 globalP->maFont = datstmpGlobals.maFont;
			 globalP->maTxtJust = datstmpGlobals.maTxtJust;
			 globalP->usePrefs = datstmpGlobals.usePrefs;
			 globalP->lineSpace = datstmpGlobals.lineSpace;
			 strcpy (globalP->textAngle, datstmpGlobals.textAngle);
			 globalP->angleVal = datstmpGlobals.angleVal;			
             mdlResource_write (globalP);
             mdlResource_free (globalP);
             }
        mdlResource_closeFile (dataUserPrefs);
        mdlState_clear ();
   }

return SUCCESS;
}

/*----------------------------------------------------------------------+
|																		|
| name		init Application Variables									|
|																		|
+----------------------------------------------------------------------*/
Private int initAppVariables ()
   {
   DatstmpGlobals   *globalP;
   RscFileHandle  dataUserPrefs;
   char cfgVal[128];
   char tmpStr[128];
   char *cfgVarExpansion = NULL;
   int i, status, chk;

   // figure out where MicroStation is saving resource files
   mdlSystem_getCfgVar (cfgVal, "MS_RSRCPATH", sizeof(cfgVal));

   //  Expands the configuration variable 
   cfgVarExpansion = mdlSystem_expandCfgVar (cfgVal);

   if (cfgVarExpansion != NULL)
       {
       sprintf (tmpStr, "%s", cfgVarExpansion);
       free (cfgVarExpansion);
       }

   if (tmpStr)
      {   
      strcat (tmpStr, "datstmp.rsc");
   chk = mdlResource_openFile (&dataUserPrefs, tmpStr, RSC_READWRITE);

	if (chk != SUCCESS)
		{
		mdlResource_createFile (tmpStr, NULL, MAVERSION);
		mdlResource_openFile (&dataUserPrefs, 
                           tmpStr, RSC_READWRITE);
		}

   if (dataUserPrefs) 
      globalP = (DatstmpGlobals *)mdlResource_load (dataUserPrefs,
                             RTYPE_globalPrefs, RSCID_globalPrefs);
   if (!globalP) {
      datstmpGlobals.maLevel = 62;
      datstmpGlobals.maWeight = 3;
      datstmpGlobals.maColor = 0;
      datstmpGlobals.maStyle = 0;
	  datstmpGlobals.maHeight = tcb->uorpersub * tcb->subpermast;
	  datstmpGlobals.maWidth = tcb->uorpersub * tcb->subpermast;
	  datstmpGlobals.maFont = 1;
	  datstmpGlobals.maTxtJust = 0;
	  datstmpGlobals.usePrefs = 0;
	  datstmpGlobals.lineSpace = tcb->uorpersub * tcb->subpermast;
	  strcpy ( datstmpGlobals.textAngle, "0" );
	  datstmpGlobals.angleVal = 0;	  
      }
   else {
      datstmpGlobals.maLevel = globalP->maLevel;
      datstmpGlobals.maWeight = globalP->maWeight;
      datstmpGlobals.maColor = globalP->maColor;
      datstmpGlobals.maStyle = globalP->maStyle;
	  datstmpGlobals.maHeight = globalP->maHeight;
	  datstmpGlobals.maWidth = globalP->maWidth;
	  datstmpGlobals.maFont = globalP->maFont;
	  datstmpGlobals.maTxtJust = globalP->maTxtJust;
	  datstmpGlobals.usePrefs = globalP->usePrefs;
	  datstmpGlobals.lineSpace = globalP->lineSpace;	 
	  strcpy ( datstmpGlobals.textAngle, globalP->textAngle );
	  datstmpGlobals.angleVal = globalP->angleVal;
      }
   if (dataUserPrefs)
      mdlResource_closeFile (dataUserPrefs);
   }

return SUCCESS;
} 

/*----------------------------------------------------------------------+
|									|
| name		main							|
|									|
+----------------------------------------------------------------------*/
int main ()
    {
	char	    *setP;
    RscFileHandle   rfHandle;

    /* --- load our command table --- */
    if (mdlParse_loadCommandTable (NULL) == NULL)
	mdlOutput_error ("Unable to load command table.");

    mdlResource_openFile (&rfHandle, NULL, FALSE);

    // set up variables that will be evaluated within C expression strings
    setP = mdlCExpression_initializeSet (VISIBILITY_DIALOG_BOX, 0, FALSE);
    mdlDialog_publishComplexVariable (setP, "datstmpglobals",
                                            "datstmpGlobals", &datstmpGlobals); 

    // publish hook function
    mdlDialog_hookPublish (sizeof(uHooks)/sizeof(DialogHookInfo), uHooks);

	initAppVariables ();

    // function to get called at unload time
    mdlSystem_setFunction (SYSTEM_UNLOAD_PROGRAM, bye);

    return SUCCESS;
    }
    